/*
 * cm11.c - State machine for talking to the X10 cm11a
 */

#define TIMEOUT_SEC 5
#define STATUS_SEC 60

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

#ifdef DOS
#include <alloc.h>
#endif

#include "cm11.h"

static char funcStr[16][16] = {
    "All off",
    "All lights on",
    "On",
    "Off",
    "Dim",
    "Bright",
    "All Lights Off",
    "Extended Code",
    "Hail Request",
    "Hail Acknowledge",
    "Preset Dim 1",
    "Preset Dim 2",
    "Extended Data",
    "Status On",
    "Status Off",
    "Status Request"
};

#define FUNC_ADDR_MASK 0x02
#define EXT_STD_MASK   0x01
#define HEADER_MASK    0x04

#define BATT_TIMER_CLEAR 0x04
#define MON_STAT_CLEAR   0x02
#define TIMER_PURGE      0x01

#define RC_RECEIVE_POLL 0x5a
#define XC_RECEIVE_GO   0xc3
#define RC_SETCLOCK_REQ 0xa5
#define XC_EPROM_START  0xfb
#define XC_RING_ENABLE  0xeb
#define XC_RING_DISABLE 0xdb
#define RC_EPROM_ADDR   0x5b
#define RC_READY        0x55
#define XC_STATUS_REQ   0x8b
#define XC_TIMER_DNLD   0x9b
#define XC_CHECKSUM_OK  0x00
#define XC_CONFUSE      0xeb  // A confusing byte for the CM11 that will
                              // with luck make it spit out something sensical 

enum receive_state {
    RS_CONFUSED,        // Not really ready.. might be doing a retransmit
    RS_READY,           /* Ready to receive new command */
    RS_CHECKSUM,        /* Waiting for checksum */
    RS_WAIT_READY,      /* Waiting for ready signal */
    RS_SIZE,            /* Waiting for size of buffer */
    RS_RECEIVE          /* Receiving known number of chars into buffer */
};

char receive_state_string[][10] = {
    "Confused",
    "Ready",
    "Checksum",
    "Wait-Ready",
    "Size",
    "Receive"
};

       
   
struct tagX10State 
{
    enum receive_state receiveState;
    BYTE checksum;
    BYTE wantCount;
    BYTE runningCount;
    int (*receiverDispatch)(struct tagX10State *state);
    BYTE buffer[128];
    int clockSet;
    time_t time;
};

struct tagTransmitEvent 
{
    int count;
    int checksumIgnore;  // How many bytes don't get included in the cs
    BYTE data[20];
    struct tagTransmitEvent *next;
};

static struct tagTransmitEvent *gblTransmitEvent = NULL;

static int gblTransmitLocked = 0;

static BYTE ByteChecksum(BYTE *buffer,int len)
/* Checksum of first <len> bytes of buffer */
{
  BYTE cs=0;
  int i;

  for (i=0; i<len; ++i) cs+=buffer[i];
  return cs;
}

static void CrackTransmission(BYTE *buffer, int count)
{
/* Function code byte -> string */
    unsigned isFunc = *buffer & FUNC_ADDR_MASK;
    unsigned isExtended = *buffer & EXT_STD_MASK;
    unsigned int dimLevel = *buffer >> 3;
    BYTE code = buffer[1];
    int i;
    
    fprintf(stderr,"Crack transmit(%d):",count);
        
    for (i = 0; i < count; i++) fprintf(stderr," %02x", buffer[i]);
    fprintf(stderr,"\n");
    
    if (!(*buffer & 0x4)) {
        // Not standard transmission
        return;
    }
    
    if (isExtended) {
        if (!isFunc || ((code & 0x0f) != 7)) {
            printf(" bogus extcode %02x %02x %02x %02x",
                    buffer[0], buffer[1], buffer[2], buffer[3]);
        } else {
            printf(" Ext-%c-%02x-%02x-%02x",
                    gHouseChr[code >> 4], buffer[2], buffer[3], buffer[4]);
        }
    } else {
        /* Not extended */
        if (isFunc) {
            printf(" %c-%u-%s", 
                   gHouseChr[code >> 4], dimLevel,
                   funcStr[code & 0x0f]);
        } else {
            printf(" %c%u-%u",
                   gHouseChr[code >> 4],
                   deviceInt[code & 0x0f],
                   dimLevel);
        }
    }
    printf("\n");
}

static int CrackSetEpromAddress(struct tagX10State *state)
{
    char buffer[32];
    unsigned long address = state->buffer[0] << 8 | state->buffer[1];
    
    LookupSymbol(address & 0xfff, buffer, sizeof(buffer));
    fprintf(stderr,"Executed eprom macro: %s\n", buffer);
    return 0;
}

static int CrackReceivePoll(struct tagX10State *state)
{
    // Interprets the receive poll data stream
    // Interfaces to external programs (in this case, it speaks)
    time_t tt;
    int i;
    BYTE *b = state->buffer;
#ifdef SPEAK
    FILE *outF = popen("say", "w");
#else
    FILE *outF = stderr;
    time(&tt);
    fprintf(outF, "Time: %s", ctime(&tt));
    fprintf(outF, "Rec Poll: ");
#endif
    for (i = 1; i < state->wantCount; i++) {
        if (b[0] & (1 << (i-1))) {
            /* function */
            switch(b[i] & 0x0f) {
            case FN_DIM:
            case FN_BRIGHT:
                /* uses extra character */
                fprintf(outF," %s %.0f%%\n",
                       funcStr[b[i] & 0x0f],
                        //houseChr[b[i] >> 4],
                       ((float)b[i+1])/2.2);
                
                i++;
                break;
            case FN_EXT_CODE:
                /* uses two extra characters */
                fprintf(outF," %s-%c-%02x-%02x-%02x",
                       funcStr[b[i] & 0x0f],
                       gHouseChr[b[i] >> 4],
                       b[i+1], b[i+2], b[i+3]);
                i+=3;
                
                break;
            default:
                fprintf(outF," %s",
                       funcStr[b[i] & 0x0f]
                        //houseChr[b[i] >> 4]
                        );
                       
                break;
            }
        } else {
            /* address */
            char adBuf[32];

            UnitToStr(b[i], adBuf, sizeof(adBuf));
            fprintf(outF," %s", adBuf);
        }
    }
#ifdef SPEAK
    pclose(outF);
#else
    fprintf(outF,"\n");
#endif
    return 0;
}

static void WriteEvent(struct tagX10State *state)
{
    WriteSerial(state->buffer, state->wantCount);
    state->time = time(NULL);
    state->receiveState = RS_CHECKSUM;
}

                
static void TransmitEvent(struct tagX10State *state)
{
    
    // Hey, we have an event in the queue
    
    struct tagTransmitEvent *tmp = gblTransmitEvent;

    CrackTransmission(tmp->data, tmp->count);
    
    memcpy(state->buffer, tmp->data, tmp->count);
    state->wantCount = tmp->count;
    state->checksum = ByteChecksum(tmp->data+tmp->checksumIgnore,
                                  tmp->count-tmp->checksumIgnore);
    
    WriteEvent(state);
    
    // Wait for it to get unlocked
    while (gblTransmitLocked) ;
    
    // Insure we don't interrupt the linker
    gblTransmitLocked = 1;
    
    gblTransmitEvent = tmp->next;
    free(tmp);
    
    // OK, let em proceed now
    gblTransmitLocked = 0;
}

static int CrackStatus(struct tagX10State *s)
{
    int batteryTimer = s->buffer[0] << 8 | s->buffer[1];
    int seconds = s->buffer[2];
    int minutes = s->buffer[3] % 60;
    int hours = s->buffer[4] * 2 + (s->buffer[3] / 60);
    int day = s->buffer[5] + ((s->buffer[6] & 0x01) << 8);
    int dayMask = s->buffer[6] & 0x7f;
    int monitoredHouseCode = s->buffer[7] >> 4;
    int firmware = s->buffer[7] & 0x0f;
    int addressedDevices = s->buffer[8] << 8 | s->buffer[9];
    int onMonitoredDevices = s->buffer[10] << 8 | s->buffer[11];
    int dimMonitoredDevices = s->buffer[12] << 8 | s->buffer[13];

    fprintf(stderr,"Status: battery=%04x time=%d:%d:%d day=%d mask=%02x\n",
            batteryTimer, hours, minutes, seconds, day, dayMask);

    fprintf(stderr,"    firmRev=%d house=%c addressed=%04x on=%04x dim=%04x\n",
            firmware, gHouseChr[monitoredHouseCode],
            addressedDevices, onMonitoredDevices, dimMonitoredDevices);
    return 0;
}

static int RequestStatus(struct tagX10State *state)
{
    if (DEBUG_CM11)
        fprintf(stderr,"Request Status\n");

    PutSerialChar(XC_STATUS_REQ);
    state->wantCount = 14;
    state->runningCount = 0;
    state->receiveState = RS_RECEIVE;
    state->receiverDispatch = CrackStatus;
    return 0;
    
}

        
static void DownloadClock(struct tagX10State *state)
/* Download system time to CM11a clock */
{
  time_t tt = time(NULL);
  struct tm *t = localtime(&tt);
 
  if (DEBUG_CM11)
      fprintf(stderr,"DownloadClock\n");

  state->buffer[0] = XC_TIMER_DNLD;
  state->buffer[1] = t->tm_sec;
  state->buffer[2] = t->tm_min + (t->tm_hour & 1) * 60;
  state->buffer[3] = t->tm_hour >> 1;
  state->buffer[4] = t->tm_yday;
  state->buffer[5] = ((t->tm_yday >> 8) << 7) | ( 1 << t->tm_wday);
  state->buffer[6] = optMonitoredHouseCode | TIMER_PURGE;

  state->wantCount = 7;

  state->checksum = ByteChecksum(state->buffer+1, 6);
  state->clockSet = 1;
  WriteEvent(state);
}
    
int IsTimeout(struct tagX10State *state, int timeout)
{
    time_t newTime = time(NULL);

    if (newTime - state->time > timeout) {
        if (DEBUG_CM11) 
            fprintf(stderr,"Timeout!\n");
        state->time = newTime;
        return 1;
    }
    return 0;
}

int HandleReadyChar(struct tagX10State *state, BYTE ch)
{
    switch (ch) {
    case RC_RECEIVE_POLL:
        DataPurge();
        PutSerialChar(XC_RECEIVE_GO);
        state->receiverDispatch = CrackReceivePoll;
        state->receiveState = RS_SIZE;
        break;
        
    case RC_EPROM_ADDR:
        state->receiverDispatch = CrackSetEpromAddress;
        state->wantCount = 2;
        state->runningCount = 0;
        state->receiveState = RS_RECEIVE;
        break;
        
    case RC_SETCLOCK_REQ:
        DataPurge();
        fprintf(stderr,"Setclock request\n");
        DownloadClock(state);
        break;
        
    case RC_READY:
        state->receiveState = RS_READY;
        break;  /* NOP */
        
    default:
        fprintf(stderr, "Unexpected char %02x in HandleReadyChar\n",
                ch);
        return 0;
    }
    return 1;
}


void Cm11ServiceLoop(int returnOnQuiet)
/* Cheesy infinite loop to wait for characters and call the state machine
   with each character...  Note, I want to change this so that it's replaced
   with some sort of interrupt-based wakeup routine
 */

{
    time_t tt;
    time_t last_tt=0;
    int firstTime = 0;
    
    struct tagX10State state;
    state.receiveState = RS_READY;
    state.clockSet = 0;
    
    do {
        if (DEBUG_CM11)
            fprintf(stderr,"State=%s\n",
                    receive_state_string[state.receiveState]);
        
        if (DataAvailable()) {
            BYTE ch = GetSerialChar();
            state.time = time(NULL);

            if (DEBUG_CM11)
                fprintf(stderr,"R%02x ", ch);

            switch (state.receiveState) {

            case RS_CHECKSUM:
                if (state.checksum == ch) {
                    PutSerialChar(XC_CHECKSUM_OK);
                    state.receiveState = RS_WAIT_READY;
                } else if (HandleReadyChar(&state, ch)) {
                    
                    fprintf(
                        stderr,
                        "(Checksum err, expect %02x, interrupted by %02x)\n",
                        state.checksum, ch);
                } else {
                    fprintf(
                        stderr,
                        "Checksum err, expect %02x, got %02x. Retransmitting...\n",
                            state.checksum, ch);
                    WriteEvent(&state);
                }
                break;
                
            case RS_READY:
            case RS_CONFUSED:
                HandleReadyChar(&state, ch);
                break;
                    
            case RS_WAIT_READY:
                if (ch == RC_READY) {
                    state.receiveState = RS_READY;
                } else if (HandleReadyChar(&state, ch)) {
                    
                    fprintf(stderr,
                            "Expecting %02x, interrupted by %02x in wait-ready state\n",
                            RC_READY, ch);
                } else {
                    fprintf(stderr,
                            "Expecting %02x, got %02x in wait-ready state. Ignoring!\n",
                            RC_READY, ch);
                }
                break;
                
            case RS_SIZE:
                if (ch < 2 || ch > 10) {
                    fprintf(stderr, "Bogus size value %02x\n", ch);
                    state.receiveState = RS_READY;
                } else {
                    state.wantCount = ch;
                    state.runningCount = 0;
                    state.receiveState = RS_RECEIVE;
                }
                break;
                
            case RS_RECEIVE:
                state.buffer[state.runningCount++] = ch;

                if (DEBUG_CM11) {
                    fprintf(stderr,"RS_RECEIVE %d,%d\n",state.runningCount,
                            state.wantCount);
                }
                
                if (state.runningCount == state.wantCount) {
                    (*state.receiverDispatch)(&state);
                    state.receiveState = RS_READY;
                }
                break;
                
            default:
                fprintf(stderr, "Unknown state\n");
            }
        } else {
            if (!firstTime) switch (state.receiveState) {
            case RS_READY:
		time(&tt);
		if (tt - last_tt > 60*60) {
		    //Print out a timestamp every hour
		    fprintf(stderr, "Time: %s", ctime(&tt));
		    last_tt = tt;
		}
                //fprintf(stderr,"RDY\n");
                
                // Got stuff in the queue?
                if (gblTransmitEvent) {
                    fprintf(stderr,"GblTransmit\n");
                    
                    if (!state.clockSet) {
                        fprintf(stderr,"Clock\n");
                        
                        DownloadClock(&state);
                    } else {
                        fprintf(stderr,"Trx\n");
                        
                        TransmitEvent(&state);
                    }
                } else if (optDownloadClock) {
                    DownloadClock(&state);
                    optDownloadClock = 0;
                //} else if (IsTimeout(&state, STATUS_SEC-5)) {
                  //RequestStatus(&state);
                }
                else if ((IsTimeout(&state, TIMEOUT_SEC) || returnOnQuiet)
                         && optTrackStatus) {
                    RequestStatus(&state);
                    if (returnOnQuiet) optTrackStatus = 0;
                }
                break;
            case RS_CHECKSUM:
                if (IsTimeout(&state, TIMEOUT_SEC)) {
                    //PutSerialChar(XC_CONFUSE);
                    state.receiveState = RS_READY;
                    break;
                }
                
            case RS_WAIT_READY:
            case RS_CONFUSED:
            case RS_SIZE:
            case RS_RECEIVE:
                if (IsTimeout(&state, TIMEOUT_SEC)) {
                    state.receiveState = RS_READY;
                    break;
                }
                break;
                
            }
            if (!gblTransmitEvent || state.receiveState != RS_READY
                || firstTime)
                DataAvailableWait(
                    state.receiveState == RS_READY  && !firstTime ?
                    STATUS_SEC : TIMEOUT_SEC);
            firstTime = 0;
        }
        
    } while (state.receiveState != RS_READY ||
             gblTransmitEvent || optTrackStatus ||
             !returnOnQuiet);
    
    // Return if all the data is processed
    // Infinite loop if no retunOnQuiet
}

static void StickItIn(struct tagTransmitEvent *new)
{
    struct tagTransmitEvent *tmp;

    // Wait for it to get unlocked
    while (gblTransmitLocked) ;
    
    // Insure we aren't re-entered
    gblTransmitLocked = 1;
    
    tmp = gblTransmitEvent;
    if (tmp == NULL) {
        gblTransmitEvent = new;
    } else {
        while (tmp->next != NULL) tmp = tmp->next;
        tmp->next = new;
    }
    // And off we go!
    gblTransmitLocked = 0;
}

void InsertSimpleEvent(BYTE header, BYTE code)
{
    struct tagTransmitEvent *new = malloc(sizeof(struct tagTransmitEvent));
    if (new == NULL) return;
    
    new->data[0] = header;
    new->data[1] = code;
    new->count = 2;
    new->next = 0;
    new->checksumIgnore = 0;
    StickItIn(new);
}

void InsertExtendedEvent(BYTE header, BYTE code, BYTE *ext)
{
    struct tagTransmitEvent *new = malloc(sizeof(struct tagTransmitEvent));
    if (new == NULL) return;
    
    new->data[0] = header;
    new->data[1] = code;
    new->data[2] = ext[0];
    new->data[3] = ext[1];
    new->data[4] = ext[2];
    new->count = 5;
    new->next = 0;
    new->checksumIgnore = 0;
//    StickItIn(new);
}

void EpromDownload(BYTE *eprom, int epromLength)
{
    int epromBase = (epromLength - 1) & 0xff0;
    int i;
    
    // First eprom event is a dummy to clear out what's in there already
    // without locking up the interface
    struct tagTransmitEvent *head = calloc(1,sizeof(struct tagTransmitEvent));
    struct tagTransmitEvent *last = head;
    
    if (!head) {
        fprintf(stderr, "Error allocating eprom stuff\n");
        return;
    }
    
    head->data[0] = XC_EPROM_START;
    for (i = 0; i< 16; i++) {
        head->data[i+3] = 0xff;
    }
    head->count = 0x13;
    head->next = 0;
    head->checksumIgnore = 1;

    if (epromLength) while (epromBase >= 0) {
        struct tagTransmitEvent *new = calloc(1,sizeof(struct tagTransmitEvent));
        if (new == NULL) {
            fprintf(stderr,"Error allocating eprom stuff.\n");
            return;
        }
        last->next = new;
        last = new;

        new->data[0] = XC_EPROM_START;
        new->data[1] = epromBase >> 8;
        new->data[2] = epromBase & 0xff;
        memcpy(new->data + 3, eprom+epromBase, 0x10);
        new->count = 0x13;
        new->next = 0;
        new->checksumIgnore = 1;
        epromBase -= 0x10;
    }
    StickItIn(head);
}


static int ImmediateLower(struct tagEvent *event) 
{
    struct tagTransmitEvent *head = NULL;
    X10_UNIT unit;
    struct tagTransmitEvent *new;
    X10_UNIT function;
    X10_UNIT_BITMAP exModMap = 0;
    
    void ClearList() {
        // Clear out all allocated memory
        if (event) free(event);
        
        while (head) {
            struct tagTransmitEvent *tmp = head->next;
            free(head);
            head = tmp;
        }
    }
    
    struct tagTransmitEvent *NewXmit() {
        // Create a new transmit event
        struct tagTransmitEvent *new =
            calloc(1, sizeof(struct tagTransmitEvent));
        if (new) {
            if (head) {
                struct tagTransmitEvent *tmp = head;
                while (tmp->next) tmp = tmp->next;
                tmp->next = new;
            } else {
                head = new;
                if (DEBUG_CM11) 
                    fprintf(stderr,"  First\n");
            }
        } else {
            fprintf(stderr, "Out of memory\n");
        }
        
        return new;
    }
    
    if (!event) return 0;
    function = event->function & 0x0f;
    
    if (DEBUG_CM11)
        fprintf(stderr,"ImmediateCommand\n");

    if (function == FN_DIM && event->dimAbsolute &&
        (exModMap = (event->deviceMask &
                     gExtModules[event->function >> 4])) != 0) {

        // Hey, we can replace some of the DIMTO's with
        // the nifty extended preset dim version
        
        if (DEBUG_CM11)
            fprintf(stderr, "Extended command %04x", exModMap);

        for (unit = 0; unit < 16; unit++) {
            if (exModMap & (1 << unit)) {
                new = NewXmit();
                if (!new) {
                    ClearList();
                    return 0;
                }
                new->count = 5;
                new->data[0] = HEADER_MASK | FUNC_ADDR_MASK | EXT_STD_MASK;
                new->data[1] = FN_EXT_CODE | (event->function & 0xf0);
                new->data[2] = unit;
                new->data[3] = ((100-event->dimLevel) * 62) / 100;
                new->data[4] = EFN_SET_BRIGHTNESS;
            }
        }
        event->deviceMask &= ~exModMap;
        if (!event->deviceMask) {
            // No more events, let's bail now
            StickItIn(head);
            
            free(event);
            return 1;
        }
    }
    
    // Now create transmit events for each unit
    for (unit = 0; unit < 16; unit++) {
        if (event->deviceMask & (1 << unit)) {

            if (DEBUG_CM11)
                fprintf(stderr," Unit %d\n",unit);
            
            new = NewXmit();
            if (!new) {
                ClearList();
                return 0;
            }
            new->count = 2;
            new->data[0] = HEADER_MASK;
            new->data[1] = unit | (event->function & 0xf0);
        }
    }

    // Now create the event for the function
    new = NewXmit();
    if (!new) {
        ClearList();
        return 0;
    }
    if (function == FN_EXT_CODE ||
        function == FN_EXT_DATA_XFER) {
        // Special case!
        new->count = 5;
        new->data[0] = HEADER_MASK | FUNC_ADDR_MASK | EXT_STD_MASK;
        new->data[1] = event->function;
        new->data[2] = event->extended[0];
        new->data[3] = event->extended[1];
        new->data[4] = event->extended[2];
    } else {
        if (event->dimAbsolute) {
            // Special case: Create an "bright" event first
            new->count = 2;
            // dim 0 = 220 steps?
            new->data[0] = HEADER_MASK | FUNC_ADDR_MASK | (22 << 3);
            new->data[1] = (event->function & 0xf0) | FN_BRIGHT;
            // And allocate another buffer
            new = NewXmit();
            if (!new) {
                ClearList();
                return 0;
            }
        } // Fall through

        new->count = 2;
        new->data[0] = HEADER_MASK | FUNC_ADDR_MASK |
            ((event->dimLevel * 22 / 100) << 3);
        new->data[1] = event->function;
    }

    // Transmit events will be freed on transmit
    StickItIn(head);
    
    free(event);
    return 1;
}

int ImmediateCommand(char * line)
{
    STREAM inStream = StringStream(line);
    struct tagEvent *event = ReadEvent(inStream);
    CloseStream(inStream);
    return ImmediateLower(event);
}

int ImmediateStream(STREAM inStream)
{
    while (!IsEndOfStream(inStream)) {
        struct tagEvent *event = ReadEvent(inStream);
        if (!ImmediateLower(event)) return 0;
    }
    return 1;
}
