/*
 * sunrise.c
 *
 * September 1994
 * Keith Walker
 *
 * This program is an adaptation of some routines from 'solar.el' that
 * is distributed with GNU emacs. As such, it is covered by the same
 * copyright as that original file:
 *
 * [This program] is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY.  No author or distributor accepts
 * responsibility to anyone for the consequences of using it or for
 * whether it serves any particular purpose or works at all, unless he
 * says so in writing.  Refer to the GNU Emacs General Public License
 * for full details.
 *
 * Everyone is granted permission to copy, modify and redistribute GNU
 * Emacs, but only under the conditions described in the GNU Emacs
 * General Public License.  A copy of this license is supposed to have
 * been given to you along with GNU Emacs so you can know your rights
 * and responsibilities.  It should be in a file named COPYING.  Among
 * other things, the copyright notice and this notice must be
 * preserved on all copies.
 *
 * The original elisp was written by Edward M. Reingold. My magic numbers
 * are his magic numbers - I don't know what they mean and don't ask; just
 * accept that they work.
 */

#define DEBUG 1


#include <stdio.h>
//#include <unistd.h>		/* for getopt */
//#include <stdlib.h>		/* for getenv */
#include <string.h>
//#include <time.h>
#include <math.h>

#include "cm11.h"
//const char *_rcs_id = "sunrise.c,v 1.4 1996/10/28 23:50:41 kew Exp kew";

#ifndef LONGITUDE
#define LONGITUDE -117.5
#endif

#ifndef LATITUDE
#define LATITUDE 47.75
#endif

double site_longitude = LONGITUDE;
double site_latitude  = LATITUDE;
double site_timezone;


const double solar_earth_orbit_eccentricity = 0.016718;
const double solar_earth_inclination = 23.441884;



double lispmod (double dividend, double divisor);
double solar_cos_inclination (void);
double solar_sin_inclination (void);
double solar_degrees_to_hours(double deg);
double solar_hours_to_days(double hours);
double degrees_to_radians (double degrees);
double radians_to_degrees (double radians);
double solar_sin_degrees (double degrees);
double solar_cosine_degrees (double degrees);
double solar_tangent_degrees (double degrees);
double solar_xy_to_quadrant (double x, double y);
double solar_degrees_to_quadrant (double angle);
double solar_arctan (double x, double quad);
double solar_arccos (double x);
double solar_arcsin (double y);
double solar_longitude_of_sun (double day);
double solar_right_ascension (double longitude);
double solar_declination (double longitude);
double solar_sunrise (int day_of_year);
double solar_sunset (int day_of_year);


double
lispmod (double f1, double f2)
{
  /* When I originally coded this up, the program didn't even come close
     to working. I tracked it down to this subtle difference between the
     fmod() routine and elisp's mod. This is an adaptation of the mod
     routine from the emacs source */
  
  f1 = fmod(f1, f2);
  if ((f1 < 0) != (f2 < 0))
    f1 += f2;
  return f1;
}


double
solar_cos_inclination (void)
{
  return solar_cosine_degrees(solar_earth_inclination);
}


double
solar_sin_inclination (void)
{
  return solar_sin_degrees(solar_earth_inclination);
}


double
solar_degrees_to_hours(double deg)
{
  /* The elisp version of this - a defmacro - always returns an
     integer (I don't understand elisp so I don't know why).  This
     causes a loss of accuracy which makes the sunrise/sunset
     calculation off by as much as 15 minutes on my machine. */
  
  return (deg / 15);
}

double
solar_hours_to_days(double hours)
{
  return hours / 24.0;
}


double
degrees_to_radians (double degrees)
{
  return degrees * (M_PI / 180.0);
}

double
radians_to_degrees (double radians)
{
  return radians * (180.0 / M_PI);
}


double
solar_sin_degrees (double degrees)
{
  return sin(degrees_to_radians(degrees));
}

double
solar_cosine_degrees (double degrees)
{
  return cos(degrees_to_radians(degrees));
}

double
solar_tangent_degrees (double degrees)
{
  return tan(degrees_to_radians(degrees));
}

double
solar_xy_to_quadrant (double x, double y)
{
  if (x > 0) {
    return (y > 0) ? 1 : 4;
  }
  else {
    return (y > 0) ? 2 : 3;
  }
}

double
solar_degrees_to_quadrant (double angle)
{
  double f = lispmod(angle, 360);
  return floor(f/90) + 1.0;
}

double
solar_arctan (double x, double quad)
{
  double deg = radians_to_degrees(atan(x));

  switch ((int)quad) {
  case 2:
  case 3:
    return deg + 180;
  case 4:
    return deg + 360;
  default:
    return deg;
  }
}

double
solar_arccos (double x)
{
  double y = sqrt(1 - (x * x));

  return solar_arctan(y / x, solar_xy_to_quadrant(x, y));
}

double
solar_arcsin (double y)
{
  double x = sqrt(1 - (y * y));

  return solar_arctan(y / x, solar_xy_to_quadrant(x, y));
}

double
solar_longitude_of_sun (double day)
{
  /* Lots of Pretty Magic Numbers! */

  double mean_anomaly = (0.9856 * day) - 3.289;

  return lispmod(mean_anomaly +
		 (1.916 * solar_sin_degrees(mean_anomaly)) +
		 (0.020 * solar_sin_degrees(2 * mean_anomaly)) + 282.634,
		 360.0);
}

double
solar_right_ascension (double longitude)
{
  double sa;
  
  sa = solar_arctan(solar_cos_inclination() * solar_tangent_degrees(longitude),
		    solar_degrees_to_quadrant(longitude));
  return solar_degrees_to_hours(sa);
}

double
solar_declination (double longitude)
{
  return solar_arcsin(solar_sin_inclination() * solar_sin_degrees(longitude));
}


double
solar_sunrise (int day_of_year)
{
//  struct tm *tm = gmtime(&date);
//  int day_of_year = tm->tm_yday;
  double approx_sunrise;
  double longitude_of_sun;
  double right_ascension;
  double declination;
  double cos_local;

  approx_sunrise = day_of_year +
    (solar_hours_to_days(6 - solar_degrees_to_hours(site_longitude)));

  longitude_of_sun = solar_longitude_of_sun(approx_sunrise);
  right_ascension = solar_right_ascension(longitude_of_sun);
  declination = solar_declination(longitude_of_sun);
  
  cos_local = ((solar_cosine_degrees(90.0 + 50.0/60.0)) -
	       ((solar_sin_degrees(declination)) *
		(solar_sin_degrees(site_latitude)))) /
		  ((solar_cosine_degrees(declination)) *
		   (solar_cosine_degrees(site_latitude)));

  if (fabs(cos_local) <= 1.0) {
    double local_sunrise;
    double local_mean_sunrise;
    double t;

    /* Hey! Even *more* magic numbers! Wow! */
    
    local_sunrise = solar_degrees_to_hours(360.0 - solar_arccos(cos_local));
    local_mean_sunrise = lispmod(((local_sunrise + right_ascension) -
			       (6.622 + (0.065710 * approx_sunrise))), 24.0);

    t = (solar_degrees_to_hours(site_longitude));
    return ((local_mean_sunrise - t) + (site_timezone / 60.0));
  }
  return 0.0;
}


double
solar_sunset (int day_of_year)
{
    //struct tm *tm = gmtime(&date);
    //int day_of_year = tm->tm_yday;
  double approx_sunset;
  double longitude_of_sun;
  double right_ascension;
  double declination;
  double cos_local;

  approx_sunset = day_of_year +
    (solar_hours_to_days(18 - solar_degrees_to_hours(site_longitude)));

  longitude_of_sun = solar_longitude_of_sun(approx_sunset);
  right_ascension = solar_right_ascension(longitude_of_sun);
  declination = solar_declination(longitude_of_sun);
  
  cos_local = ((solar_cosine_degrees(90.0 + 50.0/60.0)) -
	       ((solar_sin_degrees(declination)) *
		(solar_sin_degrees(site_latitude)))) /
		  ((solar_cosine_degrees(declination)) *
		   (solar_cosine_degrees(site_latitude)));

  if (fabs(cos_local) <= 1.0) {
    double local_sunset;
    double local_mean_sunset;
    double t;

    local_sunset = solar_degrees_to_hours(solar_arccos(cos_local));
    local_mean_sunset = lispmod(((local_sunset + right_ascension) -
			       (6.622 + (0.065710 * approx_sunset))), 24.0);

    t = (solar_degrees_to_hours(site_longitude));
    return ((local_mean_sunset - t) + (site_timezone / 60.0));
  }
  return 0.0;
}


int SunsetStrip(struct tagSunriseBin *bins, int maxBins, int toleranceMinutes)
{
    double tz = 8.;
    int i;

    int binMax;
    
    struct tagSunDay {
        double sunrise;
        double sunset;
        int next;
        int bin;
    } day[366];
    
    void RecalcAverage(int binNum) {
        double rise = 0.0;
        double set = 0.0;
        int total = 0;
        int iday = bins[binNum].lastDay;

        while (iday > -1) {
            rise += day[iday].sunrise;
            set  += day[iday].sunset;
            ++total;
            iday = day[iday].next;
        }
        bins[binNum].sunriseMinutes = (int)(rise/total);
        bins[binNum].sunsetMinutes =  (int)(set/total);
    }
    
    do {
        if (DEBUG_SUNRISE)
            fprintf(stderr,"SunsetStrip: tolerance %d minutes\n",
                    toleranceMinutes);
        binMax = 0;
        for (i=0; i < 366 && binMax < maxBins; i++) {
            day[i].sunrise = (solar_sunrise(i)-tz)*60;
            day[i].sunset = (solar_sunset(i)-tz)*60;
            
            if (i == 0) {
                day[i].next = -1;
            } else {
                if (fabs(day[i].sunrise - bins[binMax].sunriseMinutes)
                    > toleranceMinutes
                    || fabs(day[i].sunset - bins[binMax].sunsetMinutes)
                    > toleranceMinutes) {
                    ++binMax;
                    day[i].next = -1;
                } else {
                    day[i].next = i-1;
                }
            }
            if (binMax < maxBins) {
                day[i].bin = binMax;
                bins[binMax].lastDay = i;
                RecalcAverage(binMax);
            } else {
                fprintf(stderr, " Re-adjusting tolerance to fit in bin max\n");
                toleranceMinutes = 1.0 * toleranceMinutes * 370./ (i+1);
            }
        }
    } while (binMax >= maxBins);
    ++binMax;
    
    if (DEBUG_SUNRISE) {
        
        fprintf(stderr,"Bins: lastDay Sunrise, Sunset\n");
        for (i = 0; i < binMax; i++) {
            fprintf(stderr,"%d %d %d\n", bins[i].lastDay,
                    bins[i].sunriseMinutes,
                    bins[i].sunsetMinutes);
        }
        
    }
    return binMax;
}

        


